/*--------------------------------------------------------------------*/
/*  This file contains sample code.  IBM PROVIDES THIS CODE ON AN     */
/*  'AS IS' BASIS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS        */
/*  OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES  */
/*  OF MERCHANT ABILITY OR FITNESS FOR A PARTICULAR PURPOSE.          */
/*                                                                    */
/*  (C) Copyright IBM Corp. 1998, 2000                                */
/*--------------------------------------------------------------------*/
package com.ibm.etill.kitcashcassette;

import java.io.*;
import java.sql.*;
import java.util.*;

import com.ibm.etill.framework.admin.*;
import com.ibm.etill.framework.archive.*;
import com.ibm.etill.framework.cassette.*;
import com.ibm.etill.framework.log.*;
import com.ibm.etill.framework.payapi.*;
import com.ibm.etill.framework.supervisor.*;
import com.ibm.etill.kitcashcassette.test.card.*;

/**
 * <p>A <code>KitCashAccount</code> object represents an account that a merchant holds with a bank.  Merchants
 * may have one or more accounts with one or more banks.  Each account has a KitCash card that stores payments
 * made by consumers.
 * <p>KitCashAccount contains KitCashBatch. A KitCashBatch is opened implicitly if one
 * doesn't exist during addition of payments to the batch. This happens when a payment is marked for deposit.
 * A BatchClose API commands invokes the close method on the KitCashAccount object which transfers
 * the money from the Merchant to the Bank and closes the current batch.
 * <p>
 * @see com.ibm.etill.kitcashcassette.KitCashBatch
 */
public class KitCashAccount implements Archivable, KitCashAccountConstants, KitCashConstants, PaymentAPIConstants {

   // This keeps track of the KitCashAccount instances.
   private static Hashtable accounts;

   private String merchantNumber;
   private String accountName;
   private String accountNumber;
   private String bankName;
   private String bankHostName;
   private int bankPort;

   // Account contains n number of batches where n is determined by the protocol.
   // KitCash only deals with one batch at a time and hence a Vector of batches is
   // replaced with one kitCashBatch.
   private KitCashBatch kitCashBatch ;

   private FrameworkAccountAdmin accountAdmin;

   // Each KitCashAccount object has its own KitCash card to store cash.
   // KitCashDriver allows this object to access KitCash card facilities.
   private KitCashDriver driver;

   private static String sqlCreate = "INSERT INTO " + ETillArchive.geteTillDatabaseOwner() + "." + KITCASHACCOUNTS_TABLE +
                                     " ( " +
                                     COL_MERCHANTNUMBER + "," +
                                     COL_ACCOUNTNAME    + "," +
                                     COL_ACCOUNTNUMBER  + "," +
                                     COL_BANKNAME       + "," +
                                     COL_BANKHOSTNAME   + "," +
                                     COL_BANKPORT       + "," +
                                     COL_BATCHNUMBER    + ") " +
                                     "VALUES (?, ?, ?, ?, ?, ?, ?)";
   private static String sqlUpdate = "UPDATE " + ETillArchive.geteTillDatabaseOwner() + "." + KITCASHACCOUNTS_TABLE + " SET " +
                                     COL_BANKHOSTNAME + " = ?, " +
                                     COL_BANKPORT     + " = ?, " +
                                     COL_BATCHNUMBER  + " = ? " +
                                     " WHERE " + COL_MERCHANTNUMBER+ " = ? " +
                                     "   AND " + COL_ACCOUNTNUMBER + " = ? ";

   private static String sqlDelete = "DELETE FROM " + ETillArchive.geteTillDatabaseOwner() + "." + KITCASHACCOUNTS_TABLE +
                                     " WHERE " + COL_MERCHANTNUMBER+ " = ? " +
                                     "   AND " + COL_ACCOUNTNUMBER + " = ? ";

/**
 * Private ctor called only from createObject.
 * @param request com.ibm.etill.framework.cassette.AdminRequest
 * @param protocolData com.ibm.etill.framework.payapi.ParameterTable
 */
   private KitCashAccount(AdminRequest request, ParameterTable protocolData) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.KitCashAccount(AdminRequest, ParameterTable)");

      try {
         merchantNumber = (String) request.getValue(MERCHANTNUMBER);
         accountName    = (String) request.getValue(ACCOUNTTITLE);
         accountNumber  = (String) request.getValue(ACCOUNTNUMBER);
         bankName       = (String) request.getValue(FINANCIALINSTITUTION);
         bankHostName   = (String) protocolData.get(PD_BANKHOSTNAME);
         bankPort       = ((Integer) protocolData.get(PD_BANKPORT)).intValue();
         Integer batchNumber = (Integer) request.getValue(BATCHNUMBER);

         if (batchNumber != null) {
            Batch batch = Supervisor.retrieveBatch(merchantNumber, batchNumber.toString());
            kitCashBatch = (null == batch) ? null : (KitCashBatch) batch.getCassetteBatch() ;
         } else {
            kitCashBatch = null ;
         }

         accountAdmin = FrameworkAccountAdmin.obtainObject(KitCashCassette.TRACEID, merchantNumber, accountNumber);
         // Create a new KitCash card with the same amount of money on it as has
         // been collected on this Account.  (In real life, a smart card would be
         // able to remember how much money it had itself!)
         driver = new KitCashDriver(getCurrentTakings());
      } catch (Exception e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0407", e, "Failed to create a KitCashAccount Object");
         throw new ETillCassetteException("Failed to create a KitCashAccount Object");
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.KitCashAccount(AdminRequest, ParameterTable)");

   }

   /**
    * Constructs a new <code>KitCashAccount</code> object from a <code>ResultSet</code>
    * passed from <code>retrieveAccounts()</code>.
    * @param cassette not used
    * @param rs a <code>ResultSet</code> containing a KitCashAccounts database table record
    */
   private KitCashAccount(KitCashCassette cassette, ResultSet rs) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.KitCashAccount(KitCashCassette, ResultSet)");

      try {
         merchantNumber = rs.getString(COL_MERCHANTNUMBER).trim();
         if ((accountName = rs.getString(COL_ACCOUNTNAME)) != null) { // Note the asst stmt in if
            accountName = accountName.trim() ;
         }
         accountNumber  = rs.getString(COL_ACCOUNTNUMBER).trim();
         if ((bankName = rs.getString(COL_BANKNAME)) != null) { // Note the asst stmt in if
            bankName = bankName.trim();
         }
         bankHostName   = rs.getString(COL_BANKHOSTNAME).trim();
         bankPort       = rs.getInt   (COL_BANKPORT);
         int batchNumber= rs.getInt   (COL_BATCHNUMBER);

         Batch batch = Supervisor.retrieveBatch(merchantNumber, String.valueOf(batchNumber));
         kitCashBatch = (batch == null) ? null : (KitCashBatch) batch.getCassetteBatch() ;

         accountAdmin = FrameworkAccountAdmin.obtainObject(KitCashCassette.TRACEID, merchantNumber, accountNumber);
         // Create a new KitCash card with the same amount of money on it as has
         // been collected on this Account.  (In real life, a smart card would be
         // able to remember how much money it had itself!)
         driver = new KitCashDriver(getCurrentTakings());
      } catch (Exception e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0407", e, "Failed to create a KitCashAccount Object");
         throw new ETillCassetteException("Failed to create a KitCashAccount Object", cassette);
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.KitCashAccount(KitCashCassette, ResultSet)");
   }

   /**
    * Closes the batch associated with this account. Note: Real cassettes might need the
    * APIRequest as input param for retrieving the exact batch that needs to be closed.
    * @param resp the APIResponse for this transaction
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the processing fails
    */
   public void close(APIResponse resp) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.close");

      System.out.println(KitCashCassette.lineSeparator + "KitCash cassette received BATCHCLOSE command");

      // get total payment value for this batch

      int takings = getCurrentTakings();

      if (Trace.isAnyoneTracing()) {
         Trace.traceDebug(KitCashCassette.TRACEID,"-------------------------------------------------------");
         Trace.traceDebug(KitCashCassette.TRACEID,"Totals payments for batch " + kitCashBatch.getBatchNumber() + " are " + takings);
      }
      System.out.println("Totals payments for batch " + kitCashBatch.getBatchNumber() + " are " + takings);

      // process payment to bank
      boolean paymentDone = false;
      try {
         int balance = getDriver().getBalance();
         if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"Current balance: " + balance);

         // if the total payments do not equal the balance in the wallet,
         // throw an exception and let the merchant handle it.
         if (takings != balance) {
            if (Trace.isAnyoneTracing())
               Trace.traceDebug(KitCashCassette.TRACEID,"Takings ("+takings+") does not equal card balance ("+balance+")");

            ErrorLog.logError(KitCashCassette.TRACEID,"0701", null, String.valueOf(kitCashBatch.getBatchNumber()),
                              String.valueOf(takings),
                              String.valueOf(balance));
            throw new ETillAbortOperation(PRC_CASSETTE_ERROR, RC_BATCH);
         }

         // Initialize the merchant's KitCash card and tell it to send
         // a payment to the bank for the calculated amount.
         KitCashTransaction trans = getDriver().init();
         KitCashMessage msg = trans.pay(balance);

         System.out.println(KitCashCassette.lineSeparator + "KitCash protocol messages:-");
         System.out.println("-------------------------");
         while ((msg != null) && msg.toBeForwarded()) {
            // Wrap the protocol message received from the KitCash card in a MIME message
            // with the appropriate header information (bank name and account number).
            KitCashMessageWrapper request = new KitCashMessageWrapper(getBankName(),
                                                                      getAccountNumber(),
                                                                      msg);
            System.out.println(msg.toString() + " --------------------> bank");

            // Set the payment amount in the MIME header so that the bank
            // knows how much money to receive.
            request.setAmount(balance);

            // Send the MIME message containing the KitCash protocol message to the bank
            // and wait for the bank's response.
            KitCashMessageWrapper response = request.getResponse(getBankHostName(),
                                                                 getBankPort());

            // If the bank responds (which we expect it to), send the KitCash protocol
            // message that is imbedded in the bank's MIME response to the merchant's
            // KitCash card and get the next KitCash message (if any) to send back
            // to the bank.
            msg = null;
            if (response != null) {
               msg = response.getContents();
               System.out.println("merchant <-------------------- " + msg.toString());
               msg = trans.processMsg(msg);
            }
         }

         // Once the KitCash card driver stops supplying messages, the flow is complete.
         paymentDone = true;
         if ((msg != null) && msg.isComplete()) {
            if (Trace.isAnyoneTracing()) {
               Trace.traceDebug(KitCashCassette.TRACEID,"Transfer complete!");
               Trace.traceDebug(KitCashCassette.TRACEID,"Current balance: " + getDriver().getBalance());
               Trace.traceDebug(KitCashCassette.TRACEID,"-------------------------------------------------------");
            }

            kitCashBatch.close() ;
            kitCashBatch = null ; // set the kitCashBatch to null.

            // add this to the update list.
            Supervisor.getThreadCommitPoint().addToUpdateList(this);
            Supervisor.getThreadCommitPoint().commit();

         } else throw new KitCashException("No message from receiver");
      } catch (IOException ioe) {
         Supervisor.getThreadCommitPoint().noCommit();
         ErrorLog.logError(KitCashCassette.TRACEID,"0702", ioe, String.valueOf(kitCashBatch.getBatchNumber()),
                           getBankHostName(),
                           String.valueOf(getBankPort()));
         throw new ETillAbortOperation(PRC_COMMUNICATION_ERROR, RC_BATCH);
      } catch (KitCashException kce) {
         Supervisor.getThreadCommitPoint().noCommit();
         ErrorLog.logError(KitCashCassette.TRACEID,"0703", kce, String.valueOf(kitCashBatch.getBatchNumber()));
         throw new ETillAbortOperation(PRC_CASSETTE_ERROR, RC_BATCH);
      }
      resp.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.close");
   }

   /**
   *
   * Creates a new account.
   * @param request the request to create the account
   * @param protocolKeywords the protocol data for the create request
   * @param response the response to the create request
   * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
   *            if the processing fails
   *
   */
   static protected KitCashAccount createObject(AdminRequest request, ParameterTable protocolKeywords, APIResponse response) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.createObject");

      KitCashAccount newAccount = new KitCashAccount(request, protocolKeywords);
      /*****************************************
      *
      * Save new object to both in-memory list and database
      *
      *****************************************/
      String merchantNumber = (String) request.getValue(MERCHANTNUMBER);
      String accountNumber = (String) request.getValue(ACCOUNTNUMBER);

      accounts.put(merchantNumber + "/" + accountNumber, newAccount);
      Supervisor.getThreadCommitPoint().addToCreateList(newAccount);
      Supervisor.getThreadCommitPoint().commit();

      response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.createObject");

      return newAccount;
   }

   /**
    * Requests that a KitCashAccount object be created in the database.  This method is
    * required by the Archivable interface.
    * @param conn the database connection to use
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the processing fails
    */
   public void createRecord(Connection conn) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.createRecord");

      try {
         PreparedStatement pstmt = conn.prepareStatement(sqlCreate);

         int index = 1;
         pstmt.clearParameters();
         ETillArchive.addString(merchantNumber, pstmt, index++);
         ETillArchive.addString(accountName, pstmt, index++);
         ETillArchive.addString(accountNumber, pstmt, index++);
         ETillArchive.addString(bankName, pstmt, index++);
         ETillArchive.addString(bankHostName, pstmt, index++);
         ETillArchive.addInteger(new Integer(bankPort), pstmt, index++);
         Integer batchInt = new Integer(0);
         ETillArchive.addInteger(batchInt, pstmt, index++);

         if (Trace.isAnyoneTracing()) Trace.traceDatabaseWrite(KitCashCassette.TRACEID, sqlCreate);

         pstmt.executeUpdate();
         pstmt.close();
      } catch (SQLException e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0202",e,accountNumber,merchantNumber,e.toString());
         throw new ETillAbortOperation(PRC_DATABASE_ERROR, RC_NONE);
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.createRecord");

   }

   /**
   *
   * Delete the account object specified in the admin request.
   * @param request the request to delete the account
   * @param protocolKeywords the protocol data for the delete request
   * @param response the response to the delete request
   * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
   *            if the processing fails
   *
   */
   protected void deleteObject(AdminRequest request, ParameterTable protocolKeywords, APIResponse response) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.deleteObject");

      String merchantNumber = (String) request.getValue(MERCHANTNUMBER);
      String accountNumber = (String) request.getValue(ACCOUNTNUMBER);

      /*****************************************
      *
      * Remove object from both in-memory list and database
      *
      *****************************************/
      accounts.remove(merchantNumber + "/" + accountNumber);
      Supervisor.getThreadCommitPoint().addToDeleteList(this);
      Supervisor.getThreadCommitPoint().commit();

      response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.deleteObject");
   }

   /**
    * Requests that a KitCashAccount object be deleted.  This method is required by the
    * Archivable interface.
    * @param conn the database connection to use
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the processing fails
    */
   public void deleteRecord(Connection conn) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.deleteRecord");

      try {
         PreparedStatement pstmt = conn.prepareStatement(sqlDelete);

         int index = 1;
         pstmt.clearParameters();

         ETillArchive.addString(merchantNumber, pstmt, index++);
         ETillArchive.addString(accountNumber, pstmt, index++);

         if (Trace.isAnyoneTracing()) Trace.traceDatabaseWrite(KitCashCassette.TRACEID, sqlDelete);

         pstmt.executeUpdate();
         pstmt.close();
      } catch (SQLException e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0203",e,accountNumber,merchantNumber,e.toString());
         throw new ETillAbortOperation(PRC_DATABASE_ERROR, RC_NONE);
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.deleteRecord");
   }

   /**
    * Returns the account matching the input <code>merchantNumber</code> and <code>accountNumber</code>
    * or null if no match is found.
    * <p>The accounts are read into memory when the cassette starts up so any changes to the database
    * once the Payment Manager is running will not be registered until the Payment Manager restarts.
    *
    * @param merchantNumber the ID of the merchant
    * @param accountNumber the ID of the account
    * @return the KitCashAccount object with the given key fields
    * @see com.ibm.etill.kitcashcassette.KitCashAccount#retrieveAccounts
    */
   public static KitCashAccount getAccount(String merchantNumber, String accountNumber) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getAccount");

      if (Trace.isAnyoneTracing())
         Trace.traceDebug(KitCashCassette.TRACEID,"looking for account [" + merchantNumber + "/" + accountNumber + "]");

      KitCashAccount account = (KitCashAccount) accounts.get(merchantNumber + "/" + accountNumber);
      if (account != null) {
         if (Trace.isAnyoneTracing())
            Trace.traceDebug(KitCashCassette.TRACEID,"found account [" + account.toString() + "]");
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getAccount");
      return account;
   }

   /**
    * Returns the account name of this account.
    * @return the account name of this account
    */
   public String getAccountName() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getAccountName");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getAccountName");
      return accountAdmin.getAccountTitle();
   }

   /**
    * Returns the account number of this account.
    * @return the account number of this account
    */
   public String getAccountNumber() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getAccountNumber");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getAccountNumber");
      return accountAdmin.getAccountNumber();
   }

   /**
    * Returns the hostname of the bank holding this account.
    * @return the hostname of the bank holding this account
    */
   public String getBankHostName() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getBankHostName");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getBankHostName");
      return bankHostName;
   }

   /**
    * Returns the name of the bank holding this account.
    * @return the name of the bank holding this account
    */
   public String getBankName() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getBankName");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getBankName");
      return accountAdmin.getFinancialInstitutionTitle();
   }

   /**
    * Returns the port number of the bank holding this account.
    * @return the port number of the bank holding this account
    */
   public int getBankPort() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getBankPort");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getBankPort");
      return bankPort;
   }

   /**
    * Returns the KitCashBatch associated with this account.
    * <p>A KitCash account can only have one batch open at any time.
    * @return the KitCashBatch currently open for this account, if not creates a new KitCashBatch.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the processing fails
    */
   public KitCashBatch getBatch() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getBatch");

      synchronized (this) {
         if (null == kitCashBatch) {
            Batch batch = Supervisor.createBatch(merchantNumber, accountNumber, KitCashCassette.PaymentType) ;
            batch.setCurrentState(BATCH_OPEN);
            Supervisor.getThreadCommitPoint().addToCreateList(batch);
            kitCashBatch = (KitCashBatch) batch.getCassetteBatch() ;
            Supervisor.getThreadCommitPoint().addToUpdateList(this);
         }
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getBatch");
      return kitCashBatch;
   }

   /**
    * Returns the sales total for this account.
    * @return the amount of money current collected against this account
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if a database error occurs
    */
   public int getCurrentTakings() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getCurrentTakings");

      // If kitCashBatch is null, then the sales total must be zero.
      if (kitCashBatch == null) return 0;

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getCurrentTakings");

      //Otherwise, query the batch object for its totals.
      return kitCashBatch.getCurrentTakings() ;
   }

   /**
    * Returns the KitCash card driver for this account.
    * <p><code>KitCashPurchase</code> and <code>KitCashBatch</code> objects will need to
    * use this getter method to obtain the driver instance so
    * that payments to and from the merchant can be sent to the KitCash card.
    * @return KitCashDriver object for this Account
    */
   public KitCashDriver getDriver() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getDriver");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getDriver");
      return driver;
   }

   /**
    * Returns the merchant number of this account.
    * @return the merchant number of this account
    */
   public String getMerchantNumber() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.getMerchantNumber");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.getMerchaneNumber");
      return accountAdmin.getMerchantNumber();
   }


   /**
    * Returns <code>true</code> if there is one or more account registered on the database.
    * <p>If there are no accounts, then the cassette is deemed to be invalid and its
    * <code>verifyConfig</code> method will fail.
    * @return <code>true</code> if there are one or more accounts on the database; <code>false</code> otherwise.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#verifyConfig
    */
   public static boolean hasAccounts() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.hasAccounts");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.hasAccounts");
      return accounts.size() > 0;
   }

   /**
   *
   * Modify the account object specified in the admin request.
   * @param request the request to modify the account
   * @param protocolKeywords the protocol data for the modify request
   * @param response the response to the modify request
   * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
   *            if the processing fails
   *
   */
   protected void modifyObject(AdminRequest request, ParameterTable protocolData, APIResponse response) throws ETillAbortOperation
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.modifyObject");

      /*****************************************
      *
      * Retrieve incoming fields from protocol data
      *
      *****************************************/
      merchantNumber = (String) request.getValue(MERCHANTNUMBER);
      accountName    = (String) request.getValue(ACCOUNTTITLE);
      accountNumber  = (String) request.getValue(ACCOUNTNUMBER);
      bankName       = (String) request.getValue(FINANCIALINSTITUTION);
      bankHostName   = (String) protocolData.get(PD_BANKHOSTNAME);
      bankPort       = ((Integer) protocolData.get(PD_BANKPORT)).intValue();

      /*****************************************
      *
      * Commit changes
      *
      *****************************************/
      Supervisor.getThreadCommitPoint().addToUpdateList(this);

      response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.modifyObject");
   }

   /**
    * Recovers the KitCashBatches in BATCH_CLOSING state. For each batch, it marks all the
    * the purchases belonging to it as complete if not marked so earlier,  and marks the batch
    * itself as CLOSED towards the end.
    */
   public static void recoverInvalidPurchases() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.recoverInvalidPurchases");

      String sqlQuery = "SELECT DISTINCT A.MERCHANTNAME, A.ACCOUNTNUMBER, A.BATCHNUMBER FROM " +
                        ETillArchive.geteTillDatabaseOwner() + ".ETBATCH A " +
                        "WHERE A.CURRENTSTATE = " + PaymentAPIConstants.BATCH_CLOSING ;

      synchronized (ETillArchive.getArchiveLock()) {
         ETillArchive database = new ETillArchive();

         try {
            ResultSet rs = database.beginQuery(sqlQuery);
            boolean invalidRecordsFound = false;

            while (rs.next()) {
               invalidRecordsFound = true;
               String merch = rs.getString("MERCHANTNAME").trim();
               String acc   = rs.getString("ACCOUNTNUMBER").trim();
               String bnumber= rs.getString("BATCHNUMBER");

               if (Trace.isAnyoneTracing())
                  Trace.traceDebug(KitCashCassette.TRACEID,"KitCashPurchase: recoverInvalidPurchases(): account " + acc +
                                   " for merchant " + merch + " is invalid.  Attempting to recover it...");

               try {
                  KitCashBatch kitCashBatch ;
                  Batch batch = Supervisor.retrieveBatch(merch, bnumber);
                  if (batch != null) {
                     kitCashBatch = (KitCashBatch) batch.getCassetteBatch() ;
                     if (kitCashBatch == null) {
                        if (Trace.isAnyoneTracing())
                           Trace.traceDebug(KitCashCassette.TRACEID,"KitCashPurchase: recoverInvalidPurchases(): batch no longer exists! Cannot recover!");
                     } else {
                           kitCashBatch.markOrdersAsComplete() ;
                           if (Trace.isAnyoneTracing())
                              Trace.traceDebug(KitCashCassette.TRACEID,"KitCashPurchase: recoverInvalidPurchases(): account recovered successfully!");
                     }
                  }
               } catch (ETillAbortOperation eao) {
                           if (Trace.isAnyoneTracing())
                              Trace.traceDebug(KitCashCassette.TRACEID,"KitCashPurchase: recoverInvalidPurchases(): exception " + eao.toString() +
                                               " occurred recovering batch.");
               }
            }

            if (!invalidRecordsFound) {
               if (Trace.isAnyoneTracing())
                  Trace.traceDebug(KitCashCassette.TRACEID,"KitCashPurchase: recoverInvalidPurchases(): all accounts are in a valid state");
            }
         } catch (SQLException e) {
            ErrorLog.logError(KitCashCassette.TRACEID,"0982", e);
         } finally {
            try {database.endQuery();} catch (SQLException sqle) {ErrorLog.logError(KitCashCassette.TRACEID,"0469", sqle);}
         }
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.recoverInvalidPurchases");
   }

   /**
    * Reads the accounts from the KitCashAccounts database table into memory.
    * <p>This method is called when the cassette is initialized.
    * @param cassette The cassette the accounts are to be retrieved for.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if the database cannot be read.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#readCassetteConfig
    */
   public static void retrieveAccounts(KitCashCassette cassette) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.retrieveAccounts");

      accounts = new Hashtable();

      ETillArchive database = new ETillArchive();

      // Execute the given SQL statement and saving the results.
      synchronized (ETillArchive.getArchiveLock()) {
         try {
            ResultSet rs = database.beginQuery("SELECT * FROM " +
                                               ETillArchive.geteTillDatabaseOwner() + "." + KITCASHACCOUNTS_TABLE);

            while (rs.next()) {
               // See if we already have this account in the hashtable.
               // If so, don't create it or add it again.
               String merchantnumber = rs.getString(COL_MERCHANTNUMBER).trim();
               String accountnumber  = rs.getString(COL_ACCOUNTNUMBER).trim();
               KitCashAccount acc = getAccount(merchantnumber, accountnumber);
               if (acc == null) {
                  acc = new KitCashAccount(cassette, rs);

                  // store the account in a hashtable for later retrieval
                  accounts.put(acc.getMerchantNumber() + "/" + acc.getAccountNumber(), acc);
               }
               if (Trace.isAnyoneTracing())
                  Trace.traceDebug(KitCashCassette.TRACEID,"KitCashAccount: retrieved account" + KitCashCassette.lineSeparator + acc.toString());
            }
         } catch (Exception e) {
            ErrorLog.logError(KitCashCassette.TRACEID,"0408", null, e.getMessage());
            throw new ETillCassetteException("Failed to retrieve KitCashAccounts", cassette);
         } finally {
            try {database.endQuery();} catch (SQLException sqle) {ErrorLog.logError(KitCashCassette.TRACEID,"0469", sqle);}
         }
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.retrieveAccounts");
   }

   /**
    * Returns a string representation of a <code>KitCashAccounts</code> object.
    * return all member variables of a <code>KitCashAccounts</code> in a string representation.
    */
   public String toString() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.toString");

      String s = new String();

      s = COL_MERCHANTNUMBER   + " = '" + getMerchantNumber()   + "'" + KitCashCassette.lineSeparator +
          COL_ACCOUNTNAME    + " = '" + getAccountName()    + "'" + KitCashCassette.lineSeparator +
          COL_ACCOUNTNUMBER  + " = '" + getAccountNumber()  + "'" + KitCashCassette.lineSeparator +
          COL_BANKNAME       + " = '" + getBankName()       + "'" + KitCashCassette.lineSeparator +
          COL_BANKHOSTNAME   + " = '" + getBankHostName()   + "'" + KitCashCassette.lineSeparator +
          COL_BANKPORT       + " = "  + getBankPort()             + KitCashCassette.lineSeparator +
          COL_BATCHNUMBER    + " = "  + (kitCashBatch == null ? "" : kitCashBatch.getBatchNumber()) + KitCashCassette.lineSeparator +
          "Amount on KitCash card = " + driver.getBalance()       + KitCashCassette.lineSeparator;

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.toString");
      return s;
   }

   /**
    * Update this account's database record.  This method is required by the Archivable
    * interface.
    * @param conn the database connection to use
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the processing fails
    */
   public void updateRecord(Connection conn) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashAccount.updateRecord");

      try {
         PreparedStatement pstmt = conn.prepareStatement(sqlUpdate);
         int index = 1;
         pstmt.clearParameters();

         if (Trace.isAnyoneTracing()) Trace.traceDatabaseWrite(KitCashCassette.TRACEID, sqlUpdate);

         ETillArchive.addString(bankHostName, pstmt, index++);
         ETillArchive.addInteger(new Integer(bankPort), pstmt, index++);
         String batchNumber;
         if (kitCashBatch != null)
            batchNumber = kitCashBatch.getBatchNumber();
         else
            batchNumber = new String("0");
         Integer batchInt = new Integer(0);
         if (batchNumber != null)
            batchInt = Integer.valueOf(batchNumber);
         ETillArchive.addInteger(batchInt, pstmt, index++);

         ETillArchive.addString(merchantNumber, pstmt, index++);
         ETillArchive.addString(accountNumber, pstmt, index++);

         pstmt.executeUpdate();
         pstmt.close();
      } catch (SQLException e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0203",e,accountNumber,merchantNumber,e.toString());
         throw new ETillAbortOperation(PRC_DATABASE_ERROR, RC_NONE);
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashAccount.updateRecord");
   }
}
