/*--------------------------------------------------------------------*/
/*  This file contains sample code.  IBM PROVIDES THIS CODE ON AN     */
/*  'AS IS' BASIS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS        */
/*  OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES  */
/*  OF MERCHANT ABILITY OR FITNESS FOR A PARTICULAR PURPOSE.          */
/*                                                                    */
/*  (C) Copyright IBM Corp. 1998, 2000                                */
/*--------------------------------------------------------------------*/
package com.ibm.etill.kitcashcassette;

import java.io.*;
import java.sql.*;
import java.util.*;

import com.ibm.etill.framework.admin.*;
import com.ibm.etill.framework.archive.*;
import com.ibm.etill.framework.cassette.*;
import com.ibm.etill.framework.log.*;
import com.ibm.etill.framework.payapi.*;
import com.ibm.etill.framework.supervisor.*;
import com.ibm.etill.kitcashcassette.test.card.*;

/**
 * <p>A Batch is a group of transactions that are settled with a bank.
 *
 * <p>A batch is created when the Payment Manager receives a BatchOpen request.  Individual payments
 * and credits are added to or modified within the batch using the Deposit, DepositReversal, Refund
 * or RefundReversal requests.  When the Payment Manager receives a BatchClose request,
 * no further transactions may be added to the batch.
 *
 * <p><code>KitCashBatch</code> implements Batch for the KitCash cassette.
 * A <code>KitCashBatch</code> is a group a KitCash purchases for which the payment amounts have
 * not yet been deposited at a bank.  Closing a KitCashBatch performs the transfer of funds for these
 * purchases from the merchant's KitCash card to the bank.
 * <p>The KitCash payment method does not require any cassette specific Batch information to
 * be held, so it does not require a special KitCash Batch database table.
 * <p> KitCashCassette supports only receivePayment with AutoApprove and AuthDeposit. It doesn't
 * support Deposit, DepositReversal, Refund and Refund Reversal requests.
 *
 * <p>This class, along with the KitCashBatch operations like close, addPayment, markOrdersAsComplete, etc,
 * contains a constructor, getter and setter methods, methods to store and retrieve
 * Batch objects from the framework Batch table (ETBATCH) and methods to allow any Archivable objects
 * to be updated/created/deleted at a single commit point.
 *
 * @see com.ibm.etill.kitcashcassette.KitCashAccount
 */
public class KitCashBatch implements CassetteBatch, KitCashConstants, PaymentAPIConstants {

   private Batch batch;
   private Cassette cassette;

   /**
    * Constructs a new <code>KitCashBatch</code> object.
    *
    * @param cassette the cassette managing this batch object
    * @param batch the framework batch
    */
   public KitCashBatch(Cassette cassette, Batch batch) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.KitCashBatch");
      this.batch = batch;
      this.cassette  = (KitCashCassette) cassette;

      // We are going to allow merchants to close batches
      batch.setCloseAllowed(true);
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.KitCashBatch");
   }

   /**
    * Returns the cassette associated with this batch
    * @return KitCashCassette object for this batch
    */
   public Cassette cassette() {
      return cassette;
   }

   /**
    * KitCashBatch must implement createRecord from the CassetteBatch (Archivable) interface, even though
    * it does not need to provide database support.  Therefore, this method is empty.
    * @param conn the database connection to use
    */
   public void createRecord(Connection conn) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.createRecord");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.createRecord");
   }

   /**
    * KitCashBatch must implement deleteRecord from the CassetteBatch (Archivable) interface, even though
    * it does not need to provide database support.  Therefore, this method is empty.
    * @param conn the database connection to use
    */
   public void deleteRecord(Connection conn) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.deleteRecord");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.deleteRecord");
   }

   /**
    * KitCashBatch must implement updateRecord from the CassetteBatch (Archivable) interface, even though
    * it does not need to provide database support.  Therefore, this method is empty.
    * @param conn the database connection to use
    */
   public void updateRecord(Connection conn) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.updateRecord");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.updateRecord");
   }

   /**
    * Returns the sales total for this batch
    * @return the current amount of money collected against this batch
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if a database error occurs
    */
   public int getCurrentTakings() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.getCurrentTakings");

      Enumeration takingsList = batch.getTotals(); //May need to specify currency and base here.
      int takings = 0;
      BatchTotalsItem itemTotal = null;
      while (takingsList.hasMoreElements()) {
         itemTotal = (BatchTotalsItem) takingsList.nextElement();
         takings += itemTotal.getPaymentTotal();
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.getCurrentTakings");
      return takings;
   }

   /**
    * Sends a markAsComplete message to all order objects connected with this batch.
    * <p>This method is called as a result of a <code>BATCHCLOSE</code> API request.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if the processing fails
    */
   public void markOrdersAsComplete() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.markOrdersAsComplete");

      // select all orders for this batch
      Enumeration orders = batch.getRelatedOrders();

      // Send the BATCH_CLOSED_TOKEN (markAsComplete) to each order in the batch
      while (orders.hasMoreElements()) {
         String orderNumber = (String) orders.nextElement();
         Order order = Supervisor.retrieveOrder(batch.getMerchantNumber(), orderNumber);
         if (order.getCurrentState() != ORDER_CLOSED) { // if condition might be false in crash recovery case;
            KitCashPurchase purchase = (KitCashPurchase) order.getCassetteOrder();
            purchase.processEvent(new Integer(KitCashConstants.BATCH_CLOSED_TOKEN));
         }
      }

      // Once all the items in a batch are marked complete
      batch.setCurrentState(BATCH_CLOSED) ;
      batch.setTimeClosed(Supervisor.getTimestamp(System.currentTimeMillis()));

      // now remove the batch and put it in the update list.
      Supervisor.removeBatch(batch);
      Supervisor.getThreadCommitPoint().addToUpdateList(batch);
      Supervisor.getThreadCommitPoint().addToUpdateList(this);
      Supervisor.getThreadCommitPoint().commit() ;

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.markOrdersAsComplete");
   }

   /**
    * returns the batch number of the batch.
    * @return the batch number as String
    */
   public String getBatchNumber() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.getBatchNumber");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.getBatchNumber");
      return batch.getBatchNumber() ;
   }


   /**
    * closes the batch.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if close fails
    */
   public void close() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.close");

      //The account must be balanced or we wouldn't have gotten here.
      batch.setBatchStatus(PaymentAPIConstants.BATCH_BALANCED);

      batch.setCurrentState(BATCH_CLOSING);
      Supervisor.getThreadCommitPoint().addToUpdateList(batch);
      Supervisor.getThreadCommitPoint().commit() ;

      // count the number of orders associated with this batch.
      Enumeration orders = batch.getRelatedOrders();
      int numOrders = 0 ;
      for (;orders.hasMoreElements(); ++numOrders, orders.nextElement()) ;

      // if numOrders < 50 close the batch right away or else create
      // a work item and hand in over to a service thread.

      if (numOrders < 50) {
         markOrdersAsComplete() ;
      } else {
         // Create a WorkItem for marking the orders as complete, and hand it over
         // to a service thread. NOTE: If a PaymentManager Crash happens before the service
         // thread gets a chance to execute the markOrdersAsComplete, then when the payment
         // manager restarts, the batch with closing status' orders are recovered by executing
         // the markOrdersAsComplet.
         Supervisor.addItemToServiceQueue(
            new WorkItem() {
                  public void doYourWork() throws ETillAbortOperation {
                     markOrdersAsComplete();
                  }
                  public Cassette getCassette() {
                     return Admin.getCassetteAdminObject(KitCashCassette.TRACEID).getCassette();
                  }
                }                        );
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.close");
   }

   /**
    * Adds payment to the Batch object. After that, adds both the payment and batch objects to
    * commitPoint update list.
    * @param payment payment to be added to the Batch object.
    */
   public void addPayment(Payment payment) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.addPayment");
      batch.addPayment(payment) ;
      Supervisor.getThreadCommitPoint().addToUpdateList(payment);
      Supervisor.getThreadCommitPoint().addToUpdateList(batch);
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.addPayment");
   }

   /**
    * Deletes a batch from the database and memory.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation if close fails
    */
   public void delete(APIResponse apiResponse) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashBatch.delete");

      // assumes that this is called on a closed batch, in which case we don't need to
      // bother about the Account object. Account object contains a KitCashBatch
      Supervisor.removeBatch(batch) ;
      Supervisor.getThreadCommitPoint().addToDeleteList(batch) ;
      Supervisor.getThreadCommitPoint().commit() ;
      apiResponse.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashBatch.delete");
   }
}
