/*--------------------------------------------------------------------*/
/*  This file contains sample code.  IBM PROVIDES THIS CODE ON AN     */
/*  'AS IS' BASIS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS        */
/*  OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES  */
/*  OF MERCHANT ABILITY OR FITNESS FOR A PARTICULAR PURPOSE.          */
/*                                                                    */
/*  (C) Copyright IBM Corp. 1998, 2000                                */
/*--------------------------------------------------------------------*/
package com.ibm.etill.kitcashcassette;

import java.io.*;
import java.util.*;

import com.ibm.etill.framework.cassette.*;
import com.ibm.etill.framework.io.ETillConnection;
import com.ibm.etill.framework.log.*;
import com.ibm.etill.framework.payapi.*;
import com.ibm.etill.framework.supervisor.Supervisor;
import com.ibm.etill.kitcashcassette.test.card.KitCashConstants;

/**
 * <p>The Payment Manager framework creates a single instance of <code>KitCashCassette</code> at startup time.
 * This instance handles all KitCash payments for all merchants configured for the IBM Payment Manager.
 * <p>The naming convention for the cassette (including package name) is:
 * <pre>
 * com.&lt;&lt;company_name&gt;&gt;.ibmetill.&lt;&lt;lower_case_payment_system_name&gt;&gt;cassette.&lt;&lt;payment_system_name&gt;&gt;Cassette
 * </pre>
 * <p>where
 * <ul>
 * <li>&lt;&lt;company_name&gt;&gt; is in lower case and is taken from the COMPANYPKGNAME field
 * of the ETCASSETTECFG table
 * <li>&lt;&lt;payment_system_name&gt;&gt; is taken from the PAYMENTSYSTEMNAME field
 * of the ETCASSETTECFG table
 * <li>&lt;&lt;lower_case_payment_system_name&gt;&gt; is taken from the PAYMENTSYSTEMNAME field
 * of the ETCASSETTECFG table and converted to lower case.
 * </ul>
 * <p>For example, a valid cassette name would be:
 * <pre>
 * com.acme.acmecardcassette.ACMECardCassette
 * </pre>
 * <p>IBM developed cassettes use the following convention:</p>
 * <pre>
 * com.ibm.etill.framework.&lt;&lt;lower_case_payment_system_name&gt;&gt;cassette.&lt;&lt;payment_system_name&gt;&gt;Cassette
 * </pre>
 * <p>The cassette is responsible for handling startup and shutdown processing, for creating cassette-specific order,
 * payment and batch objects, and for handling requests that the framework has identified it should handle. </p>
 *
 */
public class KitCashCassette extends Cassette implements KitCashConstants, PaymentAPIConstants {

   protected static String lineSeparator = System.getProperty("line.separator");
   private static String messageID;
   KitCashComPoint comPoint;

   static String PaymentType = "KitCash";
   static String TRACEID = PaymentType;

   private Hashtable typeMapping = new Hashtable();

   /**
    * Creates all the ComPoints needed by the KitCash cassette.
    * <p>The Payment Manager framework calls <code>createComPoints()</code> when the
    * the cassette is started.
    * <p>KitCash requires only one ComPoint.  This ComPoint is used to send and receive
    * messages sent to and from KitCash wallets.
    *
    * @return a vector containing the single ComPoint needed by the KitCash cassette
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the ComPoint cannot be created and opened.
    * @see com.ibm.etill.kitcashcassette.KitCashComPoint
    */
   public Vector createComPoints() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.createComPoints");
      Vector comPoints = new Vector();

      int paymentPort = KitCashProfile.getProtocolPort();

      try {
         comPoint = new KitCashComPoint(paymentPort);
         comPoint.open();
         comPoints.addElement(comPoint);
      } catch (IOException e) {
         ErrorLog.logError(KitCashCassette.TRACEID,"0043", e);
         throw new ETillAbortOperation();
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.createComPoints");
      return comPoints;
   }
   /**
    * Creates a protocol request object constructed from the information passed
    * in the <code>ETillConnection</code>.
    * @return the requested object contructed from the ETillConnection
    * @see KitCashPaymentRequest#KitCashPaymentRequest(com.ibm.etill.framework.io.ETillConnection)
    */
   public ProtocolRequest createProtocolRequest(ETillConnection etc) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.createProtocolRequest");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.createProtocolRequest");
      return new KitCashPaymentRequest(etc);
   }
   
   /**
    * Destroys all the ComPoints that were previously created by the KitCash cassette.
    * <p>The Payment Manager framework calls <code>destroyComPoints()</code> when the
    * cassette is stopped.
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the ComPoint cannot be closed.
    * @see com.ibm.etill.kitcashcassette.KitCashComPoint
    */
   public void destroyComPoints () throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.destroyComPoints");
      try {
         comPoint.close();
      } catch (IOException e) {
         throw new ETillAbortOperation();
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.destroyComPoints");
   }
   
   /**
    * Returns the the keyword mapping hashtable.
    * <p>Payment Manager calls this method during cassette startup.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#getParameterValidationTable
    */
   public Hashtable getParameterValidationTable()
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.getParameterValidationTable");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.getParameterValidationTable");
      return typeMapping;
   }
   
   /**
    * Initializes the cassette.
    * <p>The Payment Manager framework calls this method during the cassette start-up
    * once the KitCash configuration has been read and successfully verified.
    * <p>KitCash initializes its keyword mapping hashtable in this method to make it
    * available for return when <code>getParameterValidationTable</code> is called.
    *
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the KitCash configuration is not valid.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#readCassetteConfig
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#verifyConfig
    */
   public void initializeCassette() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.initializeCassette");
      initializeKeywordMapping();
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.initializeCassette");
   }
   
   /**
    * Initializes the the keyword mapping hashtable.
    * <p>KitCash calls this method to fill in the <code>typeMaping</code> hashtable
    * for return when <code>getParameterValidationTable</code> is called by the
    * Payment Manager framework.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#getParameterValidationTable
    */
   private void initializeKeywordMapping()
   {
      /**********************************************************************
      *
      * Payment Keywords
      *
      **********************************************************************/
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.initializeKeywordMapping");
      typeMapping.put(PD_PROTOCOLPORT,  new IntegerValidationItem(PD_PROTOCOLPORT, RC_PROTOCOLPORT, false, new Integer(1), new Integer(65535)));
      typeMapping.put(PD_ACCOUNTNUMBER, new NumericTokenValidationItem(PD_ACCOUNTNUMBER, RC_ACCOUNTNUMBER));
      typeMapping.put(PD_BANKHOSTNAME,      new StringValidationItem(PD_BANKHOSTNAME, RC_BANKHOSTNAME, false, 6, 100, "8859_1"));
      typeMapping.put(PD_BANKPORT,      new IntegerValidationItem(PD_BANKPORT, RC_BANKPORT, false, new Integer(1), new Integer(65535)));
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.initializeKeywordMapping");
   }
   
   /**
    * Returns the <code>CassetteTransaction</code> object created by calling the
    * <code>KitCashBatch</code> construction.
    * <p>The Payment Manager framework will call <code>newCassetteBatch()</code> in response to a <code>BATCHOPEN</code>
    * API request.
    *
    * @param batch the framework batch object for which cassette specific information is required
    * @param The hash table containing specialized parameters for this cassette.
    * @return A new transaction object returned by <code>KitCashBatch</code>
    */
   public CassetteBatch newCassetteBatch(Batch batch, ParameterTable protocolData) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.newCassetteBatch");

      KitCashBatch newBatch = new KitCashBatch(this, batch);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.newCassetteBatch");
      return (CassetteBatch) newBatch;  
   }
   
   /**
    * Returns <code>null</code> since the KitCash payment system does not allow for credits (refunds).
    * <p>The Payment Manager framework will call <code>newCassetteCredit()</code> in response to a <code>REFUND</code>
    * API request.
    *
    * @param credit the framework credit object for which cassette specific information is required
    * @return <code>null</code>
    */
   public CassetteTransaction newCassetteCredit(Credit credit, ParameterTable protocolData) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.newCassetteCredit("+credit+")");

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.newCassetteCredit");
      return null;
   }
   
   /**
    * Creates a new <code>CassetteOrder</code> object.
    * <p>The Payment Manager framework will call <code>newCassetteOrder()</code> in response to a <code>RECEIVEPAYMENT</code>
    * or an <code>ACCEPTPAYMENT</code> API request.
    * <p>Since the KitCashPurchase object is a conglomerate of both an order and a payment,
    * this method creates the KitCashPurchase object, and <code>KitCashCassette.newCassettePayment</code>
    * returns the same object.
    *
    * @param order the framework order created by the Payment Manager framework
    * @return the new KitCash specific order
    * @see KitCashPurchase#KitCashPurchase(com.ibm.etill.framework.cassette.Cassette, com.ibm.etill.framework.payapi.Order)
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#newCassettePayment(com.ibm.etill.framework.payapi.Payment, com.ibm.etill.framework.payapi.ParameterTable)
    */
   public CassetteOrder newCassetteOrder(Order order, ParameterTable protocolData) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.newCassetteOrder("+order+")");

      KitCashPurchase purchase = new KitCashPurchase(this, order, protocolData);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.newCassetteOrder");
      return purchase;
   }
   
   /**
    * Creates a new KitCash cassette specific payment object.
    * <p>Since KitCash does not allow mutiple payments against a single order, the <code>KitCashPurchase</code>
    * object holds the cassette specific information for both the order and the payment.  This method simply
    * returns the <code>KitCashPurchase</code> object already created by <code>newCassetteOrder</code>.
    * <p>The Payment Manager framework will call <code>newCassettePayment()</code> in response to an <code>APPROVE</code>
    * API request.
    *
    * @param payment the framework order object created by the Payment Manager framework
    * @return the new KitCash specific order
    * @see KitCashPurchase#KitCashPurchase(com.ibm.etill.framework.cassette.Cassette, com.ibm.etill.framework.payapi.Order)
    */
   public CassetteTransaction newCassettePayment(Payment payment, ParameterTable protocolData) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.newCassettePayment("+payment+")");

      KitCashPurchase purchase = (KitCashPurchase) payment.getOrder().getCassetteOrder();

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.newCassettePayment");
      return purchase;
   }
   
   /**
    * This method handles admin requests from the framework.
    *
    * @param request the admin request
    * @param protocolKeywords the protocol data for the request
    * @param response the response to the request
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if an unsupported command is received.
    */
   private void processAdminRequest(AdminRequest request, ParameterTable protocolKeywords, APIResponse response) throws ETillAbortOperation
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashCassette.processAdminRequest");

      KitCashAccount account = null;

      switch (request.getToken()) {
         case MODIFY_CASSETTE_TOKEN:
            {
               if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"MODIFY_CASSETTE_TOKEN");
               Object defaultObject = new Object();
               //If there is no protocol data to process, assume this is a start operation
               //and set return code and simply return.
               if (protocolKeywords.get(PD_PROTOCOLPORT, defaultObject) == defaultObject)
                  response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);
               else
                  KitCashProfile.obtainObject().modifyObject(request, protocolKeywords, response);
            }
            break;

         case CREATE_ACCOUNT_TOKEN:
            if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"CREATE_ACCOUNT_TOKEN");
            account = KitCashAccount.getAccount((String) request.getValue(MERCHANTNUMBER), (String) request.getValue(ACCOUNTNUMBER));
            if (account != null) throw new ETillCassetteException("Fatal error - KitCash Cassette Account object already exists.");
            KitCashAccount.createObject(request, protocolKeywords, response);
            break;
         case MODIFY_ACCOUNT_TOKEN:
            if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"MODIFY_ACCOUNT_TOKEN");
            account = KitCashAccount.getAccount((String) request.getValue(MERCHANTNUMBER), (String) request.getValue(ACCOUNTNUMBER));
            if (account == null) throw new ETillCassetteException("Fatal error - KitCash Cassette Account object does not exist.");
            account.modifyObject(request, protocolKeywords, response);
            break;
         case DELETE_ACCOUNT_TOKEN:
            if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"DELETE_ACCOUNT_TOKEN");
            account = KitCashAccount.getAccount((String) request.getValue(MERCHANTNUMBER), (String) request.getValue(ACCOUNTNUMBER));
            if (account == null) throw new ETillCassetteException("Fatal error - KitCash Cassette Account object does not exist.");
            account.deleteObject(request, protocolKeywords, response);
            break;

         case CREATE_PAYSYSTEM_TOKEN:
         case MODIFY_PAYSYSTEM_TOKEN:
         case DELETE_PAYSYSTEM_TOKEN:
         case START_ACCOUNT_TOKEN:
         case STOP_ACCOUNT_TOKEN:         
         case STOP_CASSETTE_TOKEN:
         case START_PAYSYSTEM_TOKEN:
         case STOP_PAYSYSTEM_TOKEN:
            /*****************************************
            *
            * The KitCash Cassette has no duties here
            *
            *****************************************/
            if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"Ignored token");
            response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);
            break;

         case START_CASSETTE_TOKEN:
            {
               // The KitCash cassette uses this opportunity to retrieve account objects from its
               // database.
               // It also creates a KitCashCleanUp object to periodically clean up old data in the database.
               // Lastly, the cassette checks to see if there are any KitCash purchase objects in an
               // invalid state (caused by the server crashing for example) and recovers these
               // objects appropriately.
               if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"START_CASSETTE_TOKEN");
               try {
                  KitCashAccount.retrieveAccounts(this);
               } catch (ETillAbortOperation e) {
                  throw new ETillCassetteException();
               }

               // create a new database pruning work item and set it to run every 60 minutes
               KitCashCleanUp workitem = new KitCashCleanUp(60);
               workitem.doYourWork();

               // check to see if there are any KitCash purchase objects in PaymentAddedToBatch
               // that belong to a batch that has been closed.  This situation should only arise
               // if the database was forced to stop immediately after a batch had been closed
               // and before each of the purchase objects in the batch had been set to the
               // PaymentComplete state.
               KitCashAccount.recoverInvalidPurchases();
               response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);
            }
            break;

         case CASSETTECONTROL_TOKEN:
         case CREATE_MERCHANTCASSETTEOBJECT_TOKEN:
         case MODIFY_MERCHANTCASSETTEOBJECT_TOKEN:
         case DELETE_MERCHANTCASSETTEOBJECT_TOKEN:
         case CREATE_SYSTEMCASSETTEOBJECT_TOKEN:
         case MODIFY_SYSTEMCASSETTEOBJECT_TOKEN:
         case DELETE_SYSTEMCASSETTEOBJECT_TOKEN:
            throw new ETillAbortOperation(PRC_COMMAND_NOT_SUPPORTED, RC_NONE);

         default:
            throw new ETillCassetteException("Unknown Administration Command");
      }

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashProfile.processAdminRequest");
   }
   
   /**
    * Reads the KitCash configuration information from the database.  This method
    * is called by the Payment Manager framework during cassette startup.
    *
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the KitCash database tables cannot be read.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#verifyConfig
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#initializeCassette
    */
   public void readCassetteConfig() throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.readCassetteConfig");

      // Retrieve the configuration details from the KitCashConfig database table (protocol port)
      KitCashProfile.retrieveConfigRecords();
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.readCassetteConfig");
   }
   
   /**
    * Retrieves KitCash specific batch information for the provided framework batch.
    * <p>The Payment Manager framework may call this method at any time in order to retrieve the cassette specific
    * information associated with the framework batch object.
    *
    * @param batch the framework batch for which cassette specific data is required
    * @return the retrieved KitCash specific batch object
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the <code>KitCashBatch</code> object cannot be retrieved
    * @see KitCashBatch#KitCashBatch(com.ibm.etill.framework.cassette.Cassette, com.ibm.etill.framework.payapi.Batch)
    */
   public CassetteBatch resurrectCassetteBatch(Batch batch) throws ETillAbortOperation
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashCassette.resurrectCassetteBatch");
      KitCashBatch returnBatch = new KitCashBatch(this, batch);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashCassette.resurrectCassetteBatch");

      return returnBatch;
   }
   
   /**
    * Returns <code>null</code> since the KitCash payment system does not allow for credits (refunds).
    * <p>The Payment Manager framework may call this method at any time in order to retrieve the cassette specific
    * information associated with the framework credit object.
    *
    * @param credit the framework credit object for which cassette specific information is required
    * @return <code>null</code>
    */
   public CassetteTransaction resurrectCassetteCredit(Credit credit) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassetteCredit("+credit+")");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassetteCredit");

      return null;
   }
   
   /**
    * Retrieves KitCash specific order information from the KitCash database tables.
    * <p>The Payment Manager framework may call this method at any time in order to retrieve the cassette specific
    * information associated with the framework order object.
    *
    * @param order the framework order for which cassette specific data is required
    * @return the retrieved KitCash specific order object
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the <code>KitCashPurchase</code> object cannot be retrieved from the database
    * @see KitCashPurchase#KitCashPurchase(com.ibm.etill.framework.cassette.Cassette, com.ibm.etill.framework.payapi.Order)
    */
   public CassetteOrder resurrectCassetteOrder(Order order) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassetteOrder("+order+")");

      // To retrieve a KitCashPurchase object from the database, we first create a new
      // object that just points to the framework order object, then we call resurrectPurchase()
      // to go to the database and pull off the cassette specific information about the order.
      KitCashPurchase purchase = new KitCashPurchase(this, order);
      purchase.resurrectPurchase();

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassetteOrder");
      return purchase;
   }
   
   /**
    * Retrieves KitCash specific payment information from the KitCash database tables.
    * <p>Since KitCash does not allow mutiple payments against a sigle order, the <code>KitCashPurchase</code>
    * object holds the cassette specific information for both the order and the payment.  This method simply
    * returns the <code>KitCashPurchase</code> object already resurrected by <code>resurrectCassetteOrder</code>.
    * <p>The Payment Manager framework may call this method at any time in order to retrieve the cassette specific
    * information associated with the framework payment object.
    *
    * @param payment the framework payment object for which cassette specific data is required
    * @return the retrieved KitCash specific purchase object
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the <code>KitCashPurchase</code> object cannot be retrieved from the database
    * @see KitCashPurchase#KitCashPurchase(com.ibm.etill.framework.cassette.Cassette, com.ibm.etill.framework.payapi.Order)
    */
   public CassetteTransaction resurrectCassettePayment(Payment payment) throws ETillAbortOperation {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassettePayment("+payment+")");

      KitCashPurchase purchase = (KitCashPurchase) payment.getOrder().getCassetteOrder();

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.resurrectCassettePayment");
      return purchase;
   }
   
   /**
    * Services the request passed to the cassette.
    * <p>The request may be an <code>APIRequest</code> or a <code>ProtocolRequest</code>.  This method
    * will use the type of the request to determine what action to take:
    * <ul>
    * <li>Order-based payment API requests are passed to the specified KitCashPurchase object's processEvent
    * <li>KitCash protocol requests are passed to the specified KitCashPurchase object's processEvent
    * <li>A BatchClose payment API request closes the batch provided in the request.
    * <li>A DeleteBatch payment API request deletes a batch provided in the request.
    * <li>An AdminRequest is processed by the <code>processAdminRequest</code> method
    * </ul>
    *
    * @param apiRequest the API command to be performed by the cassette
    * @param apiResponse the response object
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the API command cannot be successfully processed.
    * @see KitCashPurchase#processEvent(com.ibm.etill.framework.cassette.OrderRequest, com.ibm.etill.framework.cassette.APIResponse)
    * @see KitCashPurchase#processEvent(com.ibm.etill.framework.cassette.ProtocolRequest, com.ibm.etill.framework.cassette.ProtocolResponse)
    * @see KitCashAccount#close(com.ibm.etill.framework.cassette.APIResponse)
    */
   public void service(CassetteRequest req, CassetteResponse resp) throws ETillAbortOperation {

      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashCassette.service");
      try {
         if (Trace.isAnyoneTracing()) Trace.traceDebug(KitCashCassette.TRACEID,"service(" + req + ", " + resp + ")...");

         /*****************************************
         *
         *  Process protocolKeywords
         *
         *****************************************/

         ParameterTable protocolKeywords = null;

         if (req instanceof APIRequest) {
            APIRequest request = (APIRequest) req;
            protocolKeywords = (request.getProtocolData() != null) ? request.getProtocolData() : new ParameterTable();
         } else {
            protocolKeywords = new ParameterTable();
         }

         if (req instanceof OrderRequest) {
            OrderRequest orderReq = (OrderRequest) req;
            KitCashPurchase purchase = (KitCashPurchase) (orderReq.getOrder().getCassetteOrder());

            purchase.processEvent(orderReq, (APIResponse) resp);
         } else if (req instanceof KitCashPaymentRequest) {
            KitCashPaymentRequest kcPaymentReq = (KitCashPaymentRequest) req;
            Order order = Supervisor.retrieveOrder(kcPaymentReq.getMerchantNumber(),kcPaymentReq.getOrderNumber());
            if (order == null) {
               ErrorLog.logError(KitCashCassette.TRACEID,"0138", null);
               throw new ETillAbortOperation(PRC_CASSETTE_ERROR, RC_ORDER);
            }
            KitCashPurchase purchase = (KitCashPurchase) order.getCassetteOrder();
            if (purchase == null) {
               ErrorLog.logError(KitCashCassette.TRACEID,"0101", null, kcPaymentReq.getMerchantNumber(), kcPaymentReq.getOrderNumber());
               throw new ETillAbortOperation(PRC_CASSETTE_ERROR, RC_ORDER);
            }
            if (Trace.isAnyoneTracing())
               Trace.traceDebug(KitCashCassette.TRACEID,"------> Found KitCashPurchase object");

            purchase.processEvent(kcPaymentReq, (ProtocolResponse) resp);
         } else if (req instanceof BatchCloseRequest) {
            BatchCloseRequest batchCloseReq = (BatchCloseRequest) req;
            Batch batch = batchCloseReq.getBatch(); // batch returned is non-null
            KitCashAccount account = KitCashAccount.getAccount(batch.getMerchantNumber(), batch.getAccountNumber());

            synchronized(this) {
               account.close((APIResponse) resp);
            }
         } else if (req instanceof DeleteBatchRequest) {
            DeleteBatchRequest deleteBatchRequest = (DeleteBatchRequest) req ;
            Batch batch = deleteBatchRequest.getBatch() ; // batch returned is non-null
            ((KitCashBatch) batch.getCassetteBatch()).delete((APIResponse) resp) ;
         } else if (req instanceof AdminRequest) {
            AdminRequest request = (AdminRequest) req;
            if (request.validationMode())
               validateAdminRequest(request, protocolKeywords, (APIResponse) resp);
            else
               processAdminRequest(request, protocolKeywords, (APIResponse) resp);
         } else {
            ErrorLog.logError(KitCashCassette.TRACEID,"0100", null, req.getClass().getName());
            throw new ETillAbortOperation(PRC_COMMAND_NOT_SUPPORTED,RC_NONE);
         }
      } catch (ETillRuntimeException e) {

         // The cassette is responsible for consciously choosing which exceptions should escape
         // its boundaries.  Anything other than an ETillAbortOperation that escapes the cassette
         // will cause the Framework to bring the Payment Manager down.
         // KitCashCassette consciously chooses the cases where it cannot continue, and therefore when
         // the Payment Manager should be brought down, by throwing an ETillRuntimeException.
         // If any other RuntimeException escapes from the underlying KitCashCode and is caught here,
         // it is translated into an ETillAbortOperation so that only the current operation will be aborted.
         // The Framework will continue processing new incoming requests.

         ErrorLog.logError(KitCashCassette.TRACEID,"0190",e,req.getClass().getName());
         throw e ;
      }
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashCassette.service");
   }
   
   /**
    * This method is called by the Payment Manager framework to set the unique
    * cassette name.
    * @param id the unique id assigned by the Payment Manager framework to this cassette
    */
   public void setMessageID (String id) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.setMessageID("+id+")");

      messageID = id;
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.setMessageID");
   }
   
   /**
    * Shuts down the KitCash cassette gracefully.  This method is historical.  Shutdown processing
    * should now be done in response to a STOP_CASSETTE_TOKEN in <code>processAdminRequest</code>.
    * <p>All protocol threads, service threads and timer threads allocated to this
    * cassette will be shut down by the Payment Manager.
    * <p>Since no special resources need to be freed, and no special actions need to be
    * taken before stopping the KitCash cassette, this method does nothing.
    */
   public void shutdown () {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.shutdown");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.shutdown");
   }
   
   /**************************************************************************
   * Returns whether the acceptPayment command is supported by this cassette.
   *
   * @return boolean         - true is the command is supported
   **************************************************************************/
   public boolean isAcceptPaymentSupported ()
   {
      return false;
   }   

   /**************************************************************************
   * Returns whether the receivePayment command is supported by this cassette.
   *
   * @return boolean         - true is the command is supported
   **************************************************************************/
   public boolean isReceivePaymentSupported ()
   {
      return true;
   }
   /**
    * Starts the KitCash cassette.  This method is historical.  Startup processing should
    * be done in response to the START_CASSETTE_TOKEN in <code>processAdminRequest</code>.
    * <p>The Payment Manager framework calls <code>start()</code> during cassette startup after
    * the cassette has been initialized.
    *
    * @see com.ibm.etill.kitcashcassette.KitCashCleanUp
    * @see KitCashAccount#recoverInvalidPurchases()
    */
   public void start() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashCassette.start");

      System.out.println("KitCash cassette started successfully.");

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashCassette.start");
   }

   /**
    * This method verifies that the protocol data for a particular admin request contains
    * required parameters.  Absence of required parameters causes an exception to be
    * thrown.
    *
    * @exception com.ibm.etill.framework.payapi.ETillAbortOperation
    *            if the protocol data does not contain required parameters.
    */
   private void validateAdminRequest(AdminRequest request, ParameterTable protocolKeywords, APIResponse response) throws ETillAbortOperation
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashCassette.validateAdminRequest");

      Object defaultObject = new Object();

      switch (request.getToken()) {
         case CREATE_ACCOUNT_TOKEN:
         case MODIFY_ACCOUNT_TOKEN:
            if (protocolKeywords.get(PD_BANKHOSTNAME, defaultObject) == defaultObject) throw new ETillAbortOperation(PRC_PARAMETER_NOT_FOUND, RC_BANKHOSTNAME);
            if (protocolKeywords.get(PD_BANKPORT, defaultObject) == defaultObject) throw new ETillAbortOperation(PRC_PARAMETER_NOT_FOUND, RC_BANKPORT);
            break;
      }
      response.setReturnCodes(PRC_OPERATION_SUCCESS, RC_NONE);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashCassette.validateAdminRequest");
   }
   
   /**
    * KitCashCassette requires no configuration verification, so this method
    * returns with no action.
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#readCassetteConfig
    * @see com.ibm.etill.kitcashcassette.KitCashCassette#initializeCassette
    */
   public void verifyConfig()
   {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID, "KitCashCassette.verifyConfig");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID, "KitCashCassette.verifyConfig");
   }
}
