/*--------------------------------------------------------------------*/
/*  This file contains sample code.  IBM PROVIDES THIS CODE ON AN     */
/*  'AS IS' BASIS WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS        */
/*  OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES  */
/*  OF MERCHANT ABILITY OR FITNESS FOR A PARTICULAR PURPOSE.          */
/*                                                                    */
/*  (C) Copyright IBM Corp. 1998, 2000                                */
/*--------------------------------------------------------------------*/
package com.ibm.etill.kitcashcassette;

import java.io.*;
import java.net.*;

import com.ibm.etill.framework.io.ETillConnection;
import com.ibm.etill.framework.log.*;
import com.ibm.etill.kitcashcassette.test.card.KitCashException;
import com.ibm.etill.kitcashcassette.test.card.KitCashMessage;

/**
 * <p>A <code>KitCashConnection</code> instance is created every time a consumer's wallet opens
 * a socket on a <code>KitCashComPoint</code>'s server socket. The <code>KitCashConnection</code>
 * object is passed by the Payment Manager framework to the cassette (using the
 * <code>KitCashCassette.service(protocolRequest, protocolResponse)</code> method.
 * The cassette can then use this <code>Connection</code> object to read and write messages
 * to and from the consumer's wallet.
 * <p>Most of the actual reading and writing of KitCash protocol messages is handled by the
 * <code>KitCashMessageWrapper</code> class.
 *
 * @see com.ibm.etill.kitcashcassette.KitCashMessageWrapper
 * @see com.ibm.etill.kitcashcassette.KitCashComPoint#establishConnection
 * @see com.ibm.etill.kitcashcassette.KitCashCassette#service(ProtocolRequest, ProtocolResponse)
 */
public class KitCashConnection implements ETillConnection {

   private Socket protocolSocket = null;

   /**
    * Constructs a new socket connection between wallet and cassette.
    *
    * @param socket the socket used by the KitCashComPoint
    */
   public KitCashConnection(Socket socket) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.KitCashConnection");
      protocolSocket = socket;
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.KitCashConnection");
   }

   /**
    * Closes the socket associated with this object.
    * @exception java.io.IOException if the operation fails
    */
   public void closeConnection() throws IOException {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.closeConnection");
      protocolSocket.close();
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.closeConnection");
   }

   /**
    * Reads the next KitCashMessageWrapper received at the KitCashComPoint.
    * @return a KitCashMessageWrapper object
    * @exception java.io.IOException if the operation fails
    * @see com.ibm.etill.kitcashcassette.KitCashMessageWrapper#readMessage(java.io.InputStream)
    */
   public Object read() throws IOException {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.read");
      // Get an input stream from the KitCashComPoint's socket
      DataInputStream stream = new DataInputStream(protocolSocket.getInputStream());

      // Read the message from the input stream
      KitCashMessageWrapper wrapper = KitCashMessageWrapper.readMessage(stream);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.read");
      return wrapper;
   }

   /**
    * @deprecated not required by cassette developers
    */
   public void setConnection(Object o) {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.setConnection");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.setConnection");
   }

   /**
    * Returns a string representation of the KitCashConnection object.
    * @return key information about the KitCashConnection object
    */
   public String toString() {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.toString");
      String s = getClass().getName() + KitCashCassette.lineSeparator;
      if (protocolSocket != null)
         s = s + "     Socket = " + protocolSocket.toString();
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.toString");
      return s;
   }

   /**
    * Writes the KitCashMessage parameter to the KitCashComPoint.
    * @param message the KitCashMessage to be sent
    * @return null
    * @exception java.io.IOException if the operation fails
    * @see com.ibm.etill.kitcashcassette.KitCashMessageWrapper
    * @see com.ibm.etill.kitcashcassette.test.card.KitCashMessage
    */
   public Object write(Object message) throws IOException {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.write");
      KitCashMessage msg = (KitCashMessage) message;

      // wrap the KitCash protocol message in a MIME message
      KitCashMessageWrapper wrapper = new KitCashMessageWrapper("Wallet", null, msg);

      // write the wrapped message to the socket
      DataOutputStream dos = new DataOutputStream(protocolSocket.getOutputStream());
      wrapper.writeMessage(dos);

      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.write");
      return null;
   }

   /**
    * Writes the KitCashException parameter to the KitCashComPoint.
    * @param kce the KitCashException to be sent
    * @exception java.io.IOException if the operation fails
    * @see com.ibm.etill.kitcashcassette.test.card.KitCashException
    * @see com.ibm.etill.kitcashcassette.KitCashMessageWrapper
    */
   public void writeErrorResponse(KitCashException kce) throws IOException {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.writeErrorResponse(KitCashException)");
      // wrap the KitCash error message in a MIME message
      KitCashMessageWrapper wrapper = new KitCashMessageWrapper("Wallet", null, kce);

      // write the error message to the socket
      DataOutputStream dos = new DataOutputStream(protocolSocket.getOutputStream());
      wrapper.writeMessage(dos);
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.writeErrorResponse(KitCashException)");
   }

   /**
    * @deprecated cassette developers should define their own error response methods (if required)
    */
   public Object writeErrorResponse(Object message) throws IOException {
      if (Trace.isAnyoneTracing()) Trace.traceFunctionEntry(KitCashCassette.TRACEID,"KitCashConnection.writeErrorResponse(Object)");
      if (Trace.isAnyoneTracing()) Trace.traceFunctionExit(KitCashCassette.TRACEID,"KitCashConnection.writeErrorResponse(Object)");
      return null;
   }
}
